/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg-mpi.h,v 1.9 2010/07/01 05:50:46 zlb Exp $ */

#ifndef PHG_MPI_H

extern int phgRank, phgNProcs, phgCmdSerialno;
extern GRID *_phg_g_sys;

#if USE_MPI

#define RegisterParallelFunction(f)	f(NULL)

#define ParallelFunction(f, flag)					\
    {									\
	static int func_no = -1;					\
	if (func_no < 0) {						\
	    func_no = phgRegisterParallelFunction(f, __func__,		\
			__FILE__, __LINE__);				\
	    return;							\
	}								\
	if (phgMasterSlave) {						\
	    if ((flag) && phgNProcs > 1 && phgRank == 0)		\
		phgCallSlaves(func_no);					\
	    if (!(flag) && phgRank > 0)					\
		return;							\
	}								\
    }

/*typedef GRID *(*PFunc_t)(GRID *);*/
typedef void (*PFunc_t)(GRID *);

#if DEBUG
extern char (*phgProcessorNames)[MPI_MAX_PROCESSOR_NAME + 1];
extern int _phg_comm_count, _phg_comm_count_max;
extern int (*MPI_Comm_dup_real)(MPI_Comm comm, MPI_Comm *comm_out);
extern int (*MPI_Comm_create_real)(MPI_Comm comm, MPI_Group group,
						MPI_Comm *comm_out);
extern int (*MPI_Comm_free_real)(MPI_Comm *commp);
extern int (*MPI_Comm_split_real)(MPI_Comm comm, int color, int key,
						MPI_Comm *comm_out);

#define MPI_Comm_dup(a, b) {				\
    MPI_Comm_dup_real(a, b);				\
    if (*(b) != MPI_COMM_NULL)				\
	if (++_phg_comm_count > _phg_comm_count_max)	\
	    _phg_comm_count_max = _phg_comm_count;	\
}
#define MPI_Comm_create(a, b, c) {			\
    MPI_Comm_create_real(a, b, c);			\
    if (*(c) != MPI_COMM_NULL)				\
	if (++_phg_comm_count > _phg_comm_count_max)	\
	    _phg_comm_count_max = _phg_comm_count;	\
}
#define MPI_Comm_split(a, b, c, d) {			\
    MPI_Comm_split_real(a, b, c, d);			\
    if (*(d) != MPI_COMM_NULL)				\
	if (++_phg_comm_count > _phg_comm_count_max)	\
	    _phg_comm_count_max = _phg_comm_count;	\
}
#define MPI_Comm_free(a) {				\
    if (*(a) != MPI_COMM_NULL) {			\
	MPI_Comm_free_real(a);				\
	_phg_comm_count--;				\
    }							\
    else {						\
	phgWarning("%s:%d: trying to free NULL "	\
		"communicator.\n", __FILE__, __LINE__);	\
    }							\
}
#endif	/* DEBUG */

#ifdef __cplusplus
extern "C" {
#endif
int phgRegisterParallelFunction(PFunc_t func, const char *funcname,
		const char *file, int line);
void phgInitMPI(int *argc, char ***argv);
void phgCallSlaves(int cmd);
void phgFinalizeMPI(void);
void phgSlave(void);

#ifdef __cplusplus
}
#endif

#else	/* USE_MPI */

#define ParallelFunction(f, flag)
#define RegisterParallelFunction(f)

#endif	/* USE_MPI */

#define PHG_MPI_H
#endif
